#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from datetime import datetime, timedelta

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.exceptions import AirflowException
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.cloud_composer import (
    CloudComposerCreateEnvironmentOperator,
    CloudComposerDeleteEnvironmentOperator,
    CloudComposerGetEnvironmentOperator,
    CloudComposerListEnvironmentsOperator,
    CloudComposerListImageVersionsOperator,
    CloudComposerRunAirflowCLICommandOperator,
    CloudComposerTriggerDAGRunOperator,
    CloudComposerUpdateEnvironmentOperator,
)
from airflow.providers.google.cloud.sensors.cloud_composer import (
    CloudComposerDAGRunSensor,
    CloudComposerExternalTaskSensor,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
PROJECT_NUMBER = "{{ task_instance.xcom_pull('get_project_number') }}"

DAG_ID = "composer"
REGION = "us-central1"

# [START howto_operator_composer_simple_environment]

ENVIRONMENT_ID = f"test-{DAG_ID}-{ENV_ID}".replace("_", "-")
ENVIRONMENT_ID_ASYNC = f"test-deferrable-{DAG_ID}-{ENV_ID}".replace("_", "-")

ENVIRONMENT = {
    "config": {
        "software_config": {"image_version": "composer-2-airflow-2"},
        "node_config": {
            "service_account": f"{PROJECT_NUMBER}-compute@developer.gserviceaccount.com",
        },
    },
}
# [END howto_operator_composer_simple_environment]

# [START howto_operator_composer_update_environment]
UPDATED_ENVIRONMENT = {
    "labels": {
        "label": "testing",
    }
}
UPDATE_MASK = {"paths": ["labels.label"]}
# [END howto_operator_composer_update_environment]

COMMAND = "dags list -o json --verbose"


@task(task_id="get_project_number")
def get_project_number():
    """Helper function to retrieve the number of the project based on PROJECT_ID"""
    try:
        with build("cloudresourcemanager", "v1") as service:
            response = service.projects().get(projectId=PROJECT_ID).execute()
        return response["projectNumber"]
    except HttpError as exc:
        if exc.status_code == 403:
            raise AirflowException(
                "No project found with specified name, "
                "or caller does not have permissions to read specified project"
            )
        raise exc


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "composer"],
) as dag:
    # [START howto_operator_composer_image_list]
    image_versions = CloudComposerListImageVersionsOperator(
        task_id="image_versions",
        project_id=PROJECT_ID,
        region=REGION,
    )
    # [END howto_operator_composer_image_list]

    # [START howto_operator_create_composer_environment]
    create_env = CloudComposerCreateEnvironmentOperator(
        task_id="create_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
        environment=ENVIRONMENT,
    )
    # [END howto_operator_create_composer_environment]

    # [START howto_operator_create_composer_environment_deferrable_mode]
    defer_create_env = CloudComposerCreateEnvironmentOperator(
        task_id="defer_create_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID_ASYNC,
        environment=ENVIRONMENT,
        deferrable=True,
    )
    # [END howto_operator_create_composer_environment_deferrable_mode]

    # [START howto_operator_list_composer_environments]
    list_envs = CloudComposerListEnvironmentsOperator(
        task_id="list_envs", project_id=PROJECT_ID, region=REGION
    )
    # [END howto_operator_list_composer_environments]

    # [START howto_operator_get_composer_environment]
    get_env = CloudComposerGetEnvironmentOperator(
        task_id="get_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
    )
    # [END howto_operator_get_composer_environment]

    # [START howto_operator_update_composer_environment]
    update_env = CloudComposerUpdateEnvironmentOperator(
        task_id="update_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
        update_mask=UPDATE_MASK,
        environment=UPDATED_ENVIRONMENT,
    )
    # [END howto_operator_update_composer_environment]

    # [START howto_operator_update_composer_environment_deferrable_mode]
    defer_update_env = CloudComposerUpdateEnvironmentOperator(
        task_id="defer_update_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID_ASYNC,
        update_mask=UPDATE_MASK,
        environment=UPDATED_ENVIRONMENT,
        deferrable=True,
    )
    # [END howto_operator_update_composer_environment_deferrable_mode]

    # [START howto_operator_run_airflow_cli_command]
    run_airflow_cli_cmd = CloudComposerRunAirflowCLICommandOperator(
        task_id="run_airflow_cli_cmd",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
        command=COMMAND,
    )
    # [END howto_operator_run_airflow_cli_command]

    # [START howto_operator_run_airflow_cli_command_deferrable_mode]
    defer_run_airflow_cli_cmd = CloudComposerRunAirflowCLICommandOperator(
        task_id="defer_run_airflow_cli_cmd",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID_ASYNC,
        command=COMMAND,
        deferrable=True,
    )
    # [END howto_operator_run_airflow_cli_command_deferrable_mode]

    # [START howto_sensor_dag_run]
    dag_run_sensor = CloudComposerDAGRunSensor(
        task_id="dag_run_sensor",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
        composer_dag_id="airflow_monitoring",
        allowed_states=["success"],
    )
    # [END howto_sensor_dag_run]

    # [START howto_sensor_dag_run_deferrable_mode]
    defer_dag_run_sensor = CloudComposerDAGRunSensor(
        task_id="defer_dag_run_sensor",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID_ASYNC,
        composer_dag_id="airflow_monitoring",
        allowed_states=["success"],
        deferrable=True,
    )
    # [END howto_sensor_dag_run_deferrable_mode]

    # [START howto_operator_trigger_dag_run]
    trigger_dag_run = CloudComposerTriggerDAGRunOperator(
        task_id="trigger_dag_run",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
        composer_dag_id="airflow_monitoring",
    )
    # [END howto_operator_trigger_dag_run]

    # [START howto_sensor_external_task]
    external_task_sensor = CloudComposerExternalTaskSensor(
        task_id="external_task_sensor",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
        composer_external_dag_id="airflow_monitoring",
        composer_external_task_id="echo",
        allowed_states=["success"],
        execution_range=[datetime.now() - timedelta(1), datetime.now()],
    )
    # [END howto_sensor_external_task]

    # [START howto_sensor_external_task_deferrable_mode]
    defer_external_task_sensor = CloudComposerExternalTaskSensor(
        task_id="defer_external_task_sensor",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID_ASYNC,
        composer_external_dag_id="airflow_monitoring",
        composer_external_task_id="echo",
        allowed_states=["success"],
        execution_range=[datetime.now() - timedelta(1), datetime.now()],
        deferrable=True,
    )
    # [END howto_sensor_external_task_deferrable_mode]

    # [START howto_operator_delete_composer_environment]
    delete_env = CloudComposerDeleteEnvironmentOperator(
        task_id="delete_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID,
    )
    # [END howto_operator_delete_composer_environment]
    delete_env.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_delete_composer_environment_deferrable_mode]
    defer_delete_env = CloudComposerDeleteEnvironmentOperator(
        task_id="defer_delete_env",
        project_id=PROJECT_ID,
        region=REGION,
        environment_id=ENVIRONMENT_ID_ASYNC,
        deferrable=True,
    )
    # [END howto_operator_delete_composer_environment_deferrable_mode]
    defer_delete_env.trigger_rule = TriggerRule.ALL_DONE

    chain(
        # TEST SETUP
        get_project_number(),
        image_versions,
        # TEST BODY
        [create_env, defer_create_env],
        list_envs,
        get_env,
        [update_env, defer_update_env],
        [run_airflow_cli_cmd, defer_run_airflow_cli_cmd],
        [dag_run_sensor, defer_dag_run_sensor],
        trigger_dag_run,
        [external_task_sensor, defer_external_task_sensor],
        # TEST TEARDOWN
        [delete_env, defer_delete_env],
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.hooks.workflows import WorkflowsHook
from airflow.providers.google.common.consts import CLIENT_INFO

BASE_PATH = "airflow.providers.google.cloud.hooks.workflows.{}"
LOCATION = "europe-west1"
WORKFLOW_ID = "workflow_id"
EXECUTION_ID = "execution_id"
WORKFLOW = {"aa": "bb"}
EXECUTION = {"ccc": "ddd"}
EXECUTION_NESTED = {"argument": {"project_id": "project_id", "location": "us-east1"}, "test": 1}
EXECUTION_NESTED_OUTPUT = {"argument": "{'project_id': 'project_id', 'location': 'us-east1'}", "test": 1}
PROJECT_ID = "airflow-testing"
METADATA = ()
TIMEOUT = None
RETRY = None
FILTER_ = "aaaa"
ORDER_BY = "bbb"
UPDATE_MASK = "aaa,bbb"

WORKFLOW_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}"
WORKFLOW_NAME = f"projects/{PROJECT_ID}/locations/{LOCATION}/workflows/{WORKFLOW_ID}"
EXECUTION_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}/workflows/{WORKFLOW_ID}"
EXECUTION_NAME = (
    f"projects/{PROJECT_ID}/locations/{LOCATION}/workflows/{WORKFLOW_ID}/executions/{EXECUTION_ID}"
)


def mock_init(*args, **kwargs):
    pass


class TestWorkflowsHook:
    def setup_method(self, _):
        with mock.patch(BASE_PATH.format("GoogleBaseHook.__init__"), new=mock_init):
            self.hook = WorkflowsHook(gcp_conn_id="test")

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_credentials"))
    @mock.patch(BASE_PATH.format("WorkflowsClient"))
    def test_get_workflows_client(self, mock_client, mock_get_credentials):
        self.hook.get_workflows_client()
        mock_client.assert_called_once_with(
            credentials=mock_get_credentials.return_value,
            client_info=CLIENT_INFO,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_credentials"))
    @mock.patch(BASE_PATH.format("ExecutionsClient"))
    def test_get_executions_client(self, mock_client, mock_get_credentials):
        self.hook.get_executions_client()
        mock_client.assert_called_once_with(
            credentials=mock_get_credentials.return_value,
            client_info=CLIENT_INFO,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_workflows_client"))
    def test_create_workflow(self, mock_client):
        result = self.hook.create_workflow(
            workflow=WORKFLOW,
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.create_workflow.return_value == result
        mock_client.return_value.create_workflow.assert_called_once_with(
            request=dict(workflow=WORKFLOW, workflow_id=WORKFLOW_ID, parent=WORKFLOW_PARENT),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_workflows_client"))
    def test_get_workflow(self, mock_client):
        result = self.hook.get_workflow(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.get_workflow.return_value == result
        mock_client.return_value.get_workflow.assert_called_once_with(
            request=dict(name=WORKFLOW_NAME),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_workflows_client"))
    def test_update_workflow(self, mock_client):
        result = self.hook.update_workflow(
            workflow=WORKFLOW,
            update_mask=UPDATE_MASK,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.update_workflow.return_value == result
        mock_client.return_value.update_workflow.assert_called_once_with(
            request=dict(
                workflow=WORKFLOW,
                update_mask=UPDATE_MASK,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_workflows_client"))
    def test_delete_workflow(self, mock_client):
        result = self.hook.delete_workflow(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.delete_workflow.return_value == result
        mock_client.return_value.delete_workflow.assert_called_once_with(
            request=dict(name=WORKFLOW_NAME),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_workflows_client"))
    def test_list_workflows(self, mock_client):
        result = self.hook.list_workflows(
            location=LOCATION,
            project_id=PROJECT_ID,
            filter_=FILTER_,
            order_by=ORDER_BY,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.list_workflows.return_value == result
        mock_client.return_value.list_workflows.assert_called_once_with(
            request=dict(
                parent=WORKFLOW_PARENT,
                filter=FILTER_,
                order_by=ORDER_BY,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_executions_client"))
    def test_create_execution(self, mock_client):
        result = self.hook.create_execution(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            execution=EXECUTION,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.create_execution.return_value == result
        mock_client.return_value.create_execution.assert_called_once_with(
            request=dict(
                parent=EXECUTION_PARENT,
                execution=EXECUTION,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_executions_client"))
    def test_create_execution_with_nested(self, mock_client):
        result = self.hook.create_execution(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            execution=EXECUTION_NESTED,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.create_execution.return_value == result
        mock_client.return_value.create_execution.assert_called_once_with(
            request=dict(
                parent=EXECUTION_PARENT,
                execution=EXECUTION_NESTED_OUTPUT,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_executions_client"))
    def test_get_execution(self, mock_client):
        result = self.hook.get_execution(
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.get_execution.return_value == result
        mock_client.return_value.get_execution.assert_called_once_with(
            request=dict(name=EXECUTION_NAME),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_executions_client"))
    def test_cancel_execution(self, mock_client):
        result = self.hook.cancel_execution(
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.cancel_execution.return_value == result
        mock_client.return_value.cancel_execution.assert_called_once_with(
            request=dict(name=EXECUTION_NAME),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(BASE_PATH.format("WorkflowsHook.get_executions_client"))
    def test_list_execution(self, mock_client):
        result = self.hook.list_executions(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert mock_client.return_value.list_executions.return_value == result
        mock_client.return_value.list_executions.assert_called_once_with(
            request=dict(parent=EXECUTION_PARENT),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import date
from unittest import mock

import pytest

from airflow.providers.google.common.deprecated import AirflowDeprecationAdapter, deprecated

ADAPTER_PATH = "airflow.providers.google.common.deprecated"
ADAPTER_CLASS_PATH = f"{ADAPTER_PATH}.AirflowDeprecationAdapter"


class TestAirflowDeprecationAdapter:
    @mock.patch(f"{ADAPTER_CLASS_PATH}._validate_fields")
    @mock.patch(f"{ADAPTER_CLASS_PATH}._validate_removal_release")
    @mock.patch(f"{ADAPTER_CLASS_PATH}._validate_date")
    def test_init(self, mock_validate_date, mock_validate_removal_release, mock_validate_fields):
        mock_date = mock_validate_date.return_value
        mock_release = mock_validate_removal_release.return_value

        given_date = "August 22, 2024"
        given_release = None
        adapter = AirflowDeprecationAdapter(planned_removal_date=given_date)

        mock_validate_date.assert_called_once_with(given_date)
        mock_validate_removal_release.assert_called_once_with(given_release)
        mock_validate_fields.assert_called_once_with()
        assert adapter.planned_removal_date == mock_date
        assert adapter.planned_removal_release == mock_release

    @mock.patch(f"{ADAPTER_PATH}.datetime")
    def test_validate_date(self, mock_datetime):
        value = "August 22, 2024"
        expected_date = date(2024, 8, 22)
        mock_datetime.strptime.return_value.date.return_value = expected_date

        actual_date = AirflowDeprecationAdapter._validate_date(value)

        assert actual_date == expected_date
        mock_datetime.strptime.assert_called_once_with(value, "%B %d, %Y")

    @mock.patch(f"{ADAPTER_PATH}.datetime")
    def test_validate_date_none(self, mock_datetime):
        value = None

        actual_date = AirflowDeprecationAdapter._validate_date(value)

        assert actual_date is None
        assert not mock_datetime.strptime.called

    @pytest.mark.parametrize(
        "invalid_date",
        [
            "August 55, 2024",
            "NotAugust 22, 2024",
            "2024-08-22",
            "Not a date at all",
        ],
    )
    def test_validate_date_error(self, invalid_date):
        expected_error_message = (
            f"Invalid date '{invalid_date}'. "
            f"The expected format is 'Month DD, YYYY', for example 'August 22, 2024'."
        )
        with pytest.raises(ValueError, match=expected_error_message):
            AirflowDeprecationAdapter(planned_removal_date=invalid_date)

    @pytest.mark.parametrize(
        "release_string",
        [
            "apache-airflow==1.2.3",
            "apache-airflow-providers-test==1.2.3",
        ],
    )
    def test_validate_removal_release(self, release_string):
        assert AirflowDeprecationAdapter._validate_removal_release(release_string) == release_string

    def test_validate_removal_release_none(self):
        assert AirflowDeprecationAdapter._validate_removal_release(None) is None

    @pytest.mark.parametrize(
        "release_string",
        [
            "invalid-release-string",
            "apache-airflow==2",
            "apache-airflow-providers-test==2",
        ],
    )
    def test_validate_removal_version_error(self, release_string):
        msg = (
            f"\\`{release_string}\\` must follow the format "
            f"\\'apache-airflow\\(-providers-<name>\\)==<X\\.Y\\.Z>\\'."
        )
        with pytest.raises(ValueError, match=msg):
            AirflowDeprecationAdapter._validate_removal_release(release_string)

    @pytest.mark.parametrize(
        "kwargs",
        [
            {"planned_removal_date": "August 22, 2024"},
            {"planned_removal_release": "apache-airflow-providers-test==1.2.3"},
        ],
    )
    def test_validate_fields(self, kwargs):
        adapter = AirflowDeprecationAdapter(**kwargs)
        assert adapter is not None

    def test_validate_fields_error_both_removal_date_and_release(self):
        error_message = (
            "Only one of two parameters must be set: `planned_removal_date` or 'planned_removal_release'. "
            "You specified both."
        )
        with pytest.raises(ValueError, match=error_message):
            AirflowDeprecationAdapter(
                planned_removal_date="August 22, 2024",
                planned_removal_release="apache-airflow==1.2.3",
            )

    @pytest.mark.parametrize(
        ("entity", "expected_type"),
        [
            (AirflowDeprecationAdapter, "class"),
            (AirflowDeprecationAdapter.get_deprecated_msg, "function (or method)"),
        ],
    )
    def test_entity_type(self, entity, expected_type):
        assert AirflowDeprecationAdapter.entity_type(entity) == expected_type

    @pytest.mark.parametrize(
        ("module_path", "qualified_name", "_str", "expected_path"),
        [
            ("test-module", "test-qualified-name", "test-str", "test-module.test-qualified-name"),
            ("test-module", "", "test-str", "test-module"),
            ("", "test-qualified-name", "test-str", "test-str"),
            ("", "", "test-str", "test-str"),
        ],
    )
    def test_entity_path(self, module_path, qualified_name, _str, expected_path):
        mock_entity = mock.MagicMock(
            __module__=module_path,
            __qualname__=qualified_name,
            __str__=mock.MagicMock(return_value=_str),
        )
        assert AirflowDeprecationAdapter.entity_path(mock_entity) == expected_path

    @pytest.mark.parametrize(
        ("planned_removal_date", "planned_removal_release", "expected_message"),
        [
            ("August 22, 2024", None, "after August 22, 2024"),
            (None, "apache-airflow==1.2.3", "since version apache-airflow==1.2.3"),
            (
                None,
                "apache-airflow-providers-test==1.2.3",
                "since version apache-airflow-providers-test==1.2.3",
            ),
            (None, None, "in the future"),
        ],
    )
    def test_sunset_message(self, planned_removal_date, planned_removal_release, expected_message):
        adapter = AirflowDeprecationAdapter(
            planned_removal_date=planned_removal_date,
            planned_removal_release=planned_removal_release,
        )
        assert adapter.sunset_message() == expected_message

    @pytest.mark.parametrize(
        ("use_instead", "expected_message"),
        [
            (None, "There is no replacement."),
            ("replacement", "Please use `replacement` instead."),
            ("r1, r2", "Please use `r1`, `r2` instead."),
        ],
    )
    def test_replacement_message(self, use_instead, expected_message):
        adapter = AirflowDeprecationAdapter(use_instead=use_instead)
        assert adapter.replacement_message() == expected_message

    @pytest.mark.parametrize(
        ("reason", "instructions"),
        [
            ("Test reason", "Test instructions"),
            ("Test reason", None),
            (None, "Test instructions"),
            (None, None),
        ],
    )
    @mock.patch(f"{ADAPTER_CLASS_PATH}.entity_type")
    @mock.patch(f"{ADAPTER_CLASS_PATH}.entity_path")
    @mock.patch(f"{ADAPTER_CLASS_PATH}.sunset_message")
    @mock.patch(f"{ADAPTER_CLASS_PATH}.replacement_message")
    def get_deprecated_msg(
        self,
        mock_replacement_message,
        mock_sunset_message,
        mock_entity_path,
        mock_entity_type,
        reason,
        instructions,
    ):
        replacement = mock_replacement_message.return_value
        sunset = mock_sunset_message.return_value
        entity_path = mock_entity_path.return_value
        entity_type = mock_entity_type.return_value

        expected_message = (
            f"The {entity_type} `{entity_path}` is deprecated and will be removed {sunset}. {replacement}"
        )
        if reason:
            expected_message += f" The reason is: {reason}"
        if instructions:
            expected_message += f" Instructions: {instructions}"

        mock_wrapped = mock.MagicMock()
        adapter = AirflowDeprecationAdapter(reason=mock_wrapped, instructions=instructions)

        assert adapter.get_deprecated_msg(mock.MagicMock(), mock.MagicMock()) == expected_message
        mock_entity_type.assert_called_once_with(entity=mock_wrapped)
        mock_entity_path.assert_called_once_with(entity=mock_wrapped)
        mock_sunset_message.assert_called_once_with()
        mock_replacement_message.assert_called_once_with()


@mock.patch(f"{ADAPTER_PATH}.standard_deprecated")
def test_deprecated(mock_standard_deprecated):
    mock_planned_removal_date = mock.MagicMock()
    mock_planned_removal_release = mock.MagicMock()
    mock_use_instead = mock.MagicMock()
    mock_reason = mock.MagicMock()
    mock_instructions = mock.MagicMock()
    mock_adapter_cls = mock.MagicMock()
    kwargs = {
        "planned_removal_date": mock_planned_removal_date,
        "planned_removal_release": mock_planned_removal_release,
        "use_instead": mock_use_instead,
        "reason": mock_reason,
        "instructions": mock_instructions,
        "adapter_cls": mock_adapter_cls,
    }
    extra_kwargs = {
        "test1": "test1",
        "test2": "test2",
    }
    extra_args = ["test3", "test4"]

    deprecated(*extra_args, **{**kwargs, **extra_kwargs})

    mock_standard_deprecated.assert_called_once_with(*extra_args, **{**kwargs, **extra_kwargs})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Vertex AI operators."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from google.api_core.exceptions import NotFound
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.aiplatform_v1.types import Model, model_service

from airflow.providers.google.cloud.hooks.vertex_ai.model_service import ModelServiceHook
from airflow.providers.google.cloud.links.vertex_ai import (
    VertexAIModelExportLink,
    VertexAIModelLink,
    VertexAIModelListLink,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from google.api_core.retry import Retry

    from airflow.providers.common.compat.sdk import Context


class DeleteModelOperator(GoogleCloudBaseOperator):
    """
    Deletes a Model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the Model resource to be deleted.
        Could be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}` if model
        has several versions.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("region", "model_id", "project_id", "impersonation_chain")

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.model_id = self.model_id.rpartition("@")[0] if "@" in self.model_id else self.model_id
        try:
            self.log.info("Deleting model: %s", self.model_id)
            operation = hook.delete_model(
                project_id=self.project_id,
                region=self.region,
                model=self.model_id,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            hook.wait_for_operation(timeout=self.timeout, operation=operation)
            self.log.info("Model was deleted.")
        except NotFound:
            self.log.info("The Model ID %s does not exist.", self.model_id)


class GetModelOperator(GoogleCloudBaseOperator):
    """
    Retrieves a Model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the Model resource to be retrieved.
        Could be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}` if model has
        several versions.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("region", "model_id", "project_id", "impersonation_chain")
    operator_extra_links = (VertexAIModelLink(),)

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.model_id = self.model_id.rpartition("@")[0] if "@" in self.model_id else self.model_id
        try:
            self.log.info("Retrieving model: %s", self.model_id)
            model = hook.get_model(
                project_id=self.project_id,
                region=self.region,
                model_id=self.model_id,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            self.log.info("Model found. Model ID: %s", self.model_id)

            context["ti"].xcom_push(key="model_id", value=self.model_id)
            VertexAIModelLink.persist(context=context, model_id=self.model_id)
            return Model.to_dict(model)
        except NotFound:
            self.log.info("The Model ID %s does not exist.", self.model_id)


class ExportModelOperator(GoogleCloudBaseOperator):
    """
    Exports a trained, exportable Model to a location specified by the user.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the Model to export.
        Could be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}` if model has
        several versions.
    :param output_config:  Required. The desired output location and configuration.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("region", "model_id", "project_id", "impersonation_chain")
    operator_extra_links = (VertexAIModelExportLink(),)

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        model_id: str,
        output_config: model_service.ExportModelRequest.OutputConfig | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.output_config = output_config
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        try:
            self.log.info("Exporting model: %s", self.model_id)
            operation = hook.export_model(
                project_id=self.project_id,
                region=self.region,
                model=self.model_id,
                output_config=self.output_config,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            hook.wait_for_operation(timeout=self.timeout, operation=operation)
            VertexAIModelExportLink.persist(
                context=context,
                output_config=self.output_config,
                model_id=self.model_id,
                project_id=self.project_id,
            )
            self.log.info("Model was exported.")
        except NotFound:
            self.log.info("The Model ID %s does not exist.", self.model_id)


class ListModelsOperator(GoogleCloudBaseOperator):
    r"""
    Lists Models in a Location.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param retry: Designation of what errors, if any, should be retried.
    :param filter: An expression for filtering the results of the request. For field names both
        snake_case and camelCase are supported.
        -  ``model`` supports = and !=. ``model`` represents the Model ID,  Could be in format the
        last segment of the Model's [resource name][google.cloud.aiplatform.v1.Model.name].
        -  ``display_name`` supports = and !=
        -  ``labels`` supports general map functions that is:
        --  ``labels.key=value`` - key:value equality
        --  \`labels.key:\* or labels:key - key existence
        --  A key including a space must be quoted. ``labels."a key"``.
    :param page_size: The standard list page size.
    :param page_token: The standard list page token. Typically obtained via
        [ListModelsResponse.next_page_token][google.cloud.aiplatform.v1.ListModelsResponse.next_page_token]
        of the previous
        [ModelService.ListModels][google.cloud.aiplatform.v1.ModelService.ListModels]
        call.
    :param read_mask: Mask specifying which fields to read.
    :param order_by: A comma-separated list of fields to order by, sorted in ascending order. Use "desc"
        after a field name for descending.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("region", "project_id", "impersonation_chain")
    operator_extra_links = (VertexAIModelListLink(),)

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        filter: str | None = None,
        page_size: int | None = None,
        page_token: str | None = None,
        read_mask: str | None = None,
        order_by: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.filter = filter
        self.page_size = page_size
        self.page_token = page_token
        self.read_mask = read_mask
        self.order_by = order_by
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        results = hook.list_models(
            project_id=self.project_id,
            region=self.region,
            filter=self.filter,
            page_size=self.page_size,
            page_token=self.page_token,
            read_mask=self.read_mask,
            order_by=self.order_by,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        VertexAIModelListLink.persist(context=context)
        return [Model.to_dict(result) for result in results]


class UploadModelOperator(GoogleCloudBaseOperator):
    """
    Uploads a Model artifact into Vertex AI.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model: Required. The Model to create. Creating model with the name that already
        exists leads to creating new version of existing model.
    :param parent_model: The ID of the parent model to create a new version under.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("region", "project_id", "model", "parent_model", "impersonation_chain")
    operator_extra_links = (VertexAIModelLink(),)

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        model: Model | dict,
        parent_model: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.model = model
        self.parent_model = parent_model
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Upload model")
        operation = hook.upload_model(
            project_id=self.project_id,
            region=self.region,
            model=self.model,
            parent_model=self.parent_model,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        result = hook.wait_for_operation(timeout=self.timeout, operation=operation)

        model_resp = model_service.UploadModelResponse.to_dict(result)
        model_id = hook.extract_model_id(model_resp)
        self.log.info("Model was uploaded. Model ID: %s", model_id)

        context["ti"].xcom_push(key="model_id", value=model_id)
        VertexAIModelLink.persist(context=context, model_id=model_id)
        return model_resp


class ListModelVersionsOperator(GoogleCloudBaseOperator):
    """
    Lists Model versions in a Location.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the model to list versions for.
        Could be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}` if model has
        several versions.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("model_id", "region", "project_id", "impersonation_chain")

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Retrieving versions list from model: %s", self.model_id)
        results = hook.list_model_versions(
            project_id=self.project_id,
            region=self.region,
            model_id=self.model_id,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        for result in results:
            model = Model.to_dict(result)
            self.log.info("Model name: %s;", model["name"])
            self.log.info("Model version: %s, model alias %s;", model["version_id"], model["version_aliases"])
        return [Model.to_dict(result) for result in results]


class SetDefaultVersionOnModelOperator(GoogleCloudBaseOperator):
    """
    Sets the desired Model version as Default.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the model to set as default.
        Should be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}` if model
        has several versions.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("model_id", "project_id", "impersonation_chain")
    operator_extra_links = (VertexAIModelLink(),)

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info(
            "Setting version %s as default on model %s", self.model_id.rpartition("@")[0], self.model_id
        )

        updated_model = hook.set_version_as_default(
            region=self.region,
            model_id=self.model_id,
            project_id=self.project_id,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        VertexAIModelLink.persist(context=context, model_id=self.model_id)
        return Model.to_dict(updated_model)


class AddVersionAliasesOnModelOperator(GoogleCloudBaseOperator):
    """
    Adds version aliases for the Model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the model to add version aliases for.
        Should be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}`.
    :param version_aliases: List of version aliases to be added to model version.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("model_id", "project_id", "impersonation_chain")
    operator_extra_links = (VertexAIModelLink(),)

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        version_aliases: Sequence[str],
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.version_aliases = version_aliases
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info(
            "Adding aliases %s to model version %s", self.version_aliases, self.model_id.rpartition("@")[0]
        )

        updated_model = hook.add_version_aliases(
            region=self.region,
            model_id=self.model_id,
            version_aliases=self.version_aliases,
            project_id=self.project_id,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        VertexAIModelLink.persist(context=context, model_id=self.model_id)
        return Model.to_dict(updated_model)


class DeleteVersionAliasesOnModelOperator(GoogleCloudBaseOperator):
    """
    Deletes version aliases for the Model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the model to delete version aliases from.
        Should be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}`.
    :param version_aliases: List of version aliases to be deleted from model version.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("model_id", "project_id", "impersonation_chain")
    operator_extra_links = (VertexAIModelLink(),)

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        version_aliases: Sequence[str],
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.version_aliases = version_aliases
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "region": self.region,
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info(
            "Deleting aliases %s from model version %s",
            self.version_aliases,
            self.model_id.rpartition("@")[0],
        )

        updated_model = hook.delete_version_aliases(
            region=self.region,
            model_id=self.model_id,
            version_aliases=self.version_aliases,
            project_id=self.project_id,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        VertexAIModelLink.persist(context=context, model_id=self.model_id)
        return Model.to_dict(updated_model)


class DeleteModelVersionOperator(GoogleCloudBaseOperator):
    """
    Delete Model version in a Location.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param model_id: Required. The ID of the Model in which to delete version.
        Should be in format `projects/{project}/locations/{location}/models/{model_id}@{version_id}` or
        `projects/{project}/locations/{location}/models/{model_id}@{version_alias}`
        several versions.
    :param retry: Designation of what errors, if any, should be retried.
    :param timeout: The timeout for this request.
    :param metadata: Strings which should be sent along with the request as metadata.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("model_id", "project_id", "impersonation_chain")

    def __init__(
        self,
        *,
        region: str,
        project_id: str,
        model_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.region = region
        self.project_id = project_id
        self.model_id = model_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = ModelServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        try:
            self.log.info("Deleting model version: %s", self.model_id)
            operation = hook.delete_model_version(
                project_id=self.project_id,
                region=self.region,
                model_id=self.model_id,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            hook.wait_for_operation(timeout=self.timeout, operation=operation)
            self.log.info("Model version was deleted.")
        except NotFound:
            self.log.info("The Model ID %s does not exist.", self.model_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import call, patch

import pytest

from airflow.models import Connection
from airflow.providers.google.common.hooks.discovery_api import GoogleDiscoveryApiHook


class TestGoogleDiscoveryApiHook:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id="google_test",
                conn_type="google_cloud_platform",
                host="google",
                schema="refresh_token",
                login="client_id",
                password="client_secret",
            )
        )

    @patch("airflow.providers.google.common.hooks.discovery_api.build")
    @patch("airflow.providers.google.common.hooks.discovery_api.GoogleDiscoveryApiHook._authorize")
    def test_get_conn(self, mock_authorize, mock_build):
        google_discovery_api_hook = GoogleDiscoveryApiHook(
            gcp_conn_id="google_test", api_service_name="youtube", api_version="v2"
        )

        google_discovery_api_hook.get_conn()

        mock_build.assert_called_once_with(
            serviceName=google_discovery_api_hook.api_service_name,
            version=google_discovery_api_hook.api_version,
            http=mock_authorize.return_value,
            cache_discovery=False,
        )

    @patch("airflow.providers.google.common.hooks.discovery_api.getattr")
    @patch("airflow.providers.google.common.hooks.discovery_api.GoogleDiscoveryApiHook.get_conn")
    def test_query(self, mock_get_conn, mock_getattr):
        google_discovery_api_hook = GoogleDiscoveryApiHook(
            gcp_conn_id="google_test", api_service_name="analyticsreporting", api_version="v4"
        )
        endpoint = "analyticsreporting.reports.batchGet"
        data = {
            "body": {
                "reportRequests": [
                    {
                        "viewId": "180628393",
                        "dateRanges": [{"startDate": "7daysAgo", "endDate": "today"}],
                        "metrics": [{"expression": "ga:sessions"}],
                        "dimensions": [{"name": "ga:country"}],
                    }
                ]
            }
        }
        num_retries = 1

        google_discovery_api_hook.query(endpoint, data, num_retries=num_retries)

        google_api_endpoint_name_parts = endpoint.split(".")
        mock_getattr.assert_has_calls(
            [
                call(mock_get_conn.return_value, google_api_endpoint_name_parts[1]),
                call()(),
                call(mock_getattr.return_value.return_value, google_api_endpoint_name_parts[2]),
                call()(**data),
                call()().execute(num_retries=num_retries),
            ]
        )

    @patch("airflow.providers.google.common.hooks.discovery_api.getattr")
    @patch("airflow.providers.google.common.hooks.discovery_api.GoogleDiscoveryApiHook.get_conn")
    def test_query_with_pagination(self, mock_get_conn, mock_getattr):
        google_api_conn_client_sub_call = mock_getattr.return_value.return_value
        mock_getattr.return_value.side_effect = [
            google_api_conn_client_sub_call,
            google_api_conn_client_sub_call,
            google_api_conn_client_sub_call,
            google_api_conn_client_sub_call,
            google_api_conn_client_sub_call,
            None,
        ]
        google_discovery_api_hook = GoogleDiscoveryApiHook(
            gcp_conn_id="google_test", api_service_name="analyticsreporting", api_version="v4"
        )
        endpoint = "analyticsreporting.reports.batchGet"
        data = {
            "body": {
                "reportRequests": [
                    {
                        "viewId": "180628393",
                        "dateRanges": [{"startDate": "7daysAgo", "endDate": "today"}],
                        "metrics": [{"expression": "ga:sessions"}],
                        "dimensions": [{"name": "ga:country"}],
                    }
                ]
            }
        }
        num_retries = 1

        google_discovery_api_hook.query(endpoint, data, paginate=True, num_retries=num_retries)

        api_endpoint_name_parts = endpoint.split(".")
        google_api_conn_client = mock_get_conn.return_value
        mock_getattr.assert_has_calls(
            [
                call(google_api_conn_client, api_endpoint_name_parts[1]),
                call()(),
                call(google_api_conn_client_sub_call, api_endpoint_name_parts[2]),
                call()(**data),
                call()().__bool__(),
                call()().execute(num_retries=num_retries),
                call(google_api_conn_client, api_endpoint_name_parts[1]),
                call()(),
                call(google_api_conn_client_sub_call, api_endpoint_name_parts[2] + "_next"),
                call()(google_api_conn_client_sub_call, google_api_conn_client_sub_call.execute.return_value),
                call()().__bool__(),
                call()().execute(num_retries=num_retries),
                call(google_api_conn_client, api_endpoint_name_parts[1]),
                call()(),
                call(google_api_conn_client_sub_call, api_endpoint_name_parts[2] + "_next"),
                call()(google_api_conn_client_sub_call, google_api_conn_client_sub_call.execute.return_value),
            ]
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from pathlib import Path

import structlog
from google.api_core.exceptions import NotFound

from airflow.dag_processing.bundles.base import BaseDagBundle
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook


class GCSDagBundle(BaseDagBundle):
    """
    GCS Dag bundle - exposes a directory in GCS as a Dag bundle.

    This allows Airflow to load Dags directly from a GCS bucket.

    :param gcp_conn_id: Airflow connection ID for GCS.  Defaults to GoogleBaseHook.default_conn_name.
    :param bucket_name: The name of the GCS bucket containing the Dag files.
    :param prefix:  Optional subdirectory within the GCS bucket where the Dags are stored.
                    If None, Dags are assumed to be at the root of the bucket (Optional).
    """

    supports_versioning = False

    def __init__(
        self,
        *,
        gcp_conn_id: str = GoogleBaseHook.default_conn_name,
        bucket_name: str,
        prefix: str = "",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.gcp_conn_id = gcp_conn_id
        self.bucket_name = bucket_name
        self.prefix = prefix
        # Local path where GCS Dags are downloaded
        self.gcs_dags_dir: Path = self.base_dir

        log = structlog.get_logger(__name__)
        self._log = log.bind(
            bundle_name=self.name,
            version=self.version,
            bucket_name=self.bucket_name,
            prefix=self.prefix,
            gcp_conn_id=self.gcp_conn_id,
        )
        self._gcs_hook: GCSHook | None = None

    def _initialize(self):
        with self.lock():
            if not self.gcs_dags_dir.exists():
                self._log.info("Creating local Dags directory: %s", self.gcs_dags_dir)
                os.makedirs(self.gcs_dags_dir)

            if not self.gcs_dags_dir.is_dir():
                raise NotADirectoryError(f"Local Dags path: {self.gcs_dags_dir} is not a directory.")

            try:
                self.gcs_hook.get_bucket(bucket_name=self.bucket_name)
            except NotFound:
                raise ValueError(f"GCS bucket '{self.bucket_name}' does not exist.")

            if self.prefix:
                # don't check when prefix is ""
                if not self.gcs_hook.list(bucket_name=self.bucket_name, prefix=self.prefix):
                    raise ValueError(f"GCS prefix 'gs://{self.bucket_name}/{self.prefix}' does not exist.")
            self.refresh()

    def initialize(self) -> None:
        self._initialize()
        super().initialize()

    @property
    def gcs_hook(self):
        if self._gcs_hook is None:
            try:
                self._gcs_hook: GCSHook = GCSHook(gcp_conn_id=self.gcp_conn_id)  # Initialize GCS hook.
            except AirflowException as e:
                self._log.warning("Could not create GCSHook for connection %s: %s", self.gcp_conn_id, e)
        return self._gcs_hook

    def __repr__(self):
        return (
            f"<GCSDagBundle("
            f"name={self.name!r}, "
            f"bucket_name={self.bucket_name!r}, "
            f"prefix={self.prefix!r}, "
            f"version={self.version!r}"
            f")>"
        )

    def get_current_version(self) -> str | None:
        """Return the current version of the Dag bundle. Currently not supported."""
        return None

    @property
    def path(self) -> Path:
        """Return the local path to the Dag files."""
        return self.gcs_dags_dir  # Path where Dags are downloaded.

    def refresh(self) -> None:
        """Refresh the Dag bundle by re-downloading the Dags from GCS."""
        if self.version:
            raise ValueError("Refreshing a specific version is not supported")

        with self.lock():
            self._log.debug(
                "Downloading Dags from gs://%s/%s to %s", self.bucket_name, self.prefix, self.gcs_dags_dir
            )
            self.gcs_hook.sync_to_local_dir(
                bucket_name=self.bucket_name,
                prefix=self.prefix,
                local_dir=self.gcs_dags_dir,
                delete_stale=True,
            )

    def view_url(self, version: str | None = None) -> str | None:
        """
        Return a URL for viewing the Dags in GCS. Currently, versioning is not supported.

        This method is deprecated and will be removed when the minimum supported Airflow version is 3.1.
        Use `view_url_template` instead.
        """
        return self.view_url_template()

    def view_url_template(self) -> str | None:
        """Return a URL for viewing the Dags in GCS. Currently, versioning is not supported."""
        if self.version:
            raise ValueError("GCS url with version is not supported")
        if hasattr(self, "_view_url_template") and self._view_url_template:
            # Because we use this method in the view_url method, we need to handle
            # backward compatibility for Airflow versions that doesn't have the
            # _view_url_template attribute. Should be removed when we drop support for Airflow 3.0
            return self._view_url_template
        # https://console.cloud.google.com/storage/browser/<bucket-name>/<prefix>
        url = f"https://console.cloud.google.com/storage/browser/{self.bucket_name}"
        if self.prefix:
            url += f"/{self.prefix}"

        return url



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import httplib2
import pytest
from googleapiclient.errors import HttpError

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.dataflow import (
    DEFAULT_DATAFLOW_LOCATION,
    DataflowJobStatus,
)
from airflow.providers.google.cloud.operators.dataflow import (
    DataflowCreatePipelineOperator,
    DataflowDeletePipelineOperator,
    DataflowRunPipelineOperator,
    DataflowStartFlexTemplateOperator,
    DataflowStartYamlJobOperator,
    DataflowStopJobOperator,
    DataflowTemplatedJobStartOperator,
)
from airflow.version import version

TASK_ID = "test-dataflow-operator"
JOB_ID = "test-dataflow-pipeline-id"
JOB_NAME = "test-dataflow-pipeline-name"
TEMPLATE = "gs://dataflow-templates/wordcount/template_file"
PARAMETERS = {
    "inputFile": "gs://dataflow-samples/shakespeare/kinglear.txt",
    "output": "gs://test/output/my_output",
}
PY_FILE = "gs://my-bucket/my-object.py"
PY_INTERPRETER = "python3"
JAR_FILE = "gs://my-bucket/example/test.jar"
LOCAL_JAR_FILE = "/mnt/dev/example/test.jar"
JOB_CLASS = "com.test.NotMain"
PY_OPTIONS = ["-m"]
DEFAULT_OPTIONS_PYTHON = DEFAULT_OPTIONS_JAVA = {
    "project": "test",
    "stagingLocation": "gs://test/staging",
}
DEFAULT_OPTIONS_TEMPLATE = {
    "project": "test",
    "stagingLocation": "gs://test/staging",
    "tempLocation": "gs://test/temp",
    "zone": "us-central1-f",
}
ADDITIONAL_OPTIONS = {"output": "gs://test/output", "labels": {"foo": "bar"}}
TEST_VERSION = f"v{version.replace('.', '-').replace('+', '-')}"
EXPECTED_ADDITIONAL_OPTIONS = {
    "output": "gs://test/output",
    "labels": {"foo": "bar", "airflow-version": TEST_VERSION},
}
POLL_SLEEP = 30
GCS_HOOK_STRING = "airflow.providers.google.cloud.operators.dataflow.{}"
TEST_FLEX_PARAMETERS = {
    "containerSpecGcsPath": "gs://test-bucket/test-file",
    "jobName": "test-job-name",
    "parameters": {
        "inputSubscription": "test-subscription",
        "outputTable": "test-project:test-dataset.streaming_beam_sql",
    },
}
TEST_LOCATION = "custom-location"
TEST_REGION = "custom-region"
TEST_PROJECT = "test-project"
TEST_SQL_JOB_NAME = "test-sql-job-name"
TEST_DATASET = "test-dataset"
TEST_SQL_OPTIONS = {
    "bigquery-project": TEST_PROJECT,
    "bigquery-dataset": TEST_DATASET,
    "bigquery-table": "beam_output",
    "bigquery-write-disposition": "write-truncate",
}
TEST_SQL_QUERY = """
SELECT
    sales_region as sales_region,
    count(state_id) as count_state
FROM
    bigquery.table.test-project.beam_samples.beam_table
GROUP BY sales_region;
"""
TEST_SQL_JOB = {"id": "test-job-id"}
GCP_CONN_ID = "test_gcp_conn_id"
IMPERSONATION_CHAIN = ["impersonate", "this"]
CANCEL_TIMEOUT = 10 * 420
DATAFLOW_PATH = "airflow.providers.google.cloud.operators.dataflow"

TEST_PIPELINE_NAME = "test_data_pipeline_name"
TEST_PIPELINE_BODY = {
    "name": f"projects/test-datapipeline-operators/locations/test-location/pipelines/{TEST_PIPELINE_NAME}",
    "type": "PIPELINE_TYPE_BATCH",
    "workload": {
        "dataflowFlexTemplateRequest": {
            "launchParameter": {
                "containerSpecGcsPath": "gs://dataflow-templates-us-central1/latest/Word_Count_metadata",
                "jobName": "test-job",
                "environment": {"tempLocation": "test-temp-location"},
                "parameters": {
                    "inputFile": "gs://dataflow-samples/shakespeare/kinglear.txt",
                    "output": "gs://test/output/my_output",
                },
            },
            "projectId": f"{TEST_PROJECT}",
            "location": f"{TEST_LOCATION}",
        }
    },
}
CONFLICTING_DEFERABLE_WAIT_UNTIL_FINISHED = "Conflict between deferrable and wait_until_finished parameters because it makes operator as blocking when it requires to be deferred. It should be True as deferrable parameter or True as wait_until_finished."


class TestDataflowTemplatedJobStartOperator:
    @pytest.fixture
    def sync_operator(self):
        return DataflowTemplatedJobStartOperator(
            project_id=TEST_PROJECT,
            task_id=TASK_ID,
            template=TEMPLATE,
            job_name=JOB_NAME,
            parameters=PARAMETERS,
            options=DEFAULT_OPTIONS_TEMPLATE,
            dataflow_default_options={"EXTRA_OPTION": "TEST_A"},
            poll_sleep=POLL_SLEEP,
            location=TEST_LOCATION,
            environment={"maxWorkers": 2},
        )

    @pytest.fixture
    def deferrable_operator(self):
        return DataflowTemplatedJobStartOperator(
            project_id=TEST_PROJECT,
            task_id=TASK_ID,
            template=TEMPLATE,
            job_name=JOB_NAME,
            parameters=PARAMETERS,
            options=DEFAULT_OPTIONS_TEMPLATE,
            dataflow_default_options={"EXTRA_OPTION": "TEST_A"},
            poll_sleep=POLL_SLEEP,
            location=TEST_LOCATION,
            environment={"maxWorkers": 2},
            deferrable=True,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cancel_timeout=CANCEL_TIMEOUT,
        )

    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute(self, hook_mock, sync_operator):
        start_template_hook = hook_mock.return_value.start_template_dataflow
        mock_context = {"task_instance": mock.MagicMock()}
        sync_operator.execute(mock_context)
        assert hook_mock.called
        expected_options = {
            "project": "test",
            "stagingLocation": "gs://test/staging",
            "tempLocation": "gs://test/temp",
            "zone": "us-central1-f",
            "EXTRA_OPTION": "TEST_A",
        }
        start_template_hook.assert_called_once_with(
            job_name=JOB_NAME,
            variables=expected_options,
            parameters=PARAMETERS,
            dataflow_template=TEMPLATE,
            on_new_job_callback=mock.ANY,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            environment={"maxWorkers": 2},
            append_job_name=True,
        )

    @mock.patch(f"{DATAFLOW_PATH}.DataflowTemplatedJobStartOperator.defer")
    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute_with_deferrable_mode(self, mock_hook, mock_defer_method, deferrable_operator):
        deferrable_operator.execute(mock.MagicMock())
        expected_variables = {
            "project": "test",
            "stagingLocation": "gs://test/staging",
            "tempLocation": "gs://test/temp",
            "zone": "us-central1-f",
            "EXTRA_OPTION": "TEST_A",
        }
        mock_hook.return_value.launch_job_with_template.assert_called_once_with(
            job_name=JOB_NAME,
            variables=expected_variables,
            parameters=PARAMETERS,
            dataflow_template=TEMPLATE,
            project_id=TEST_PROJECT,
            append_job_name=True,
            location=TEST_LOCATION,
            environment={"maxWorkers": 2},
        )
        mock_defer_method.assert_called_once()

    def test_validation_deferrable_params_raises_error(self):
        init_kwargs = {
            "project_id": TEST_PROJECT,
            "task_id": TASK_ID,
            "template": TEMPLATE,
            "job_name": JOB_NAME,
            "parameters": PARAMETERS,
            "options": DEFAULT_OPTIONS_TEMPLATE,
            "dataflow_default_options": {"EXTRA_OPTION": "TEST_A"},
            "poll_sleep": POLL_SLEEP,
            "location": TEST_LOCATION,
            "environment": {"maxWorkers": 2},
            "wait_until_finished": True,
            "deferrable": True,
            "gcp_conn_id": GCP_CONN_ID,
            "impersonation_chain": IMPERSONATION_CHAIN,
            "cancel_timeout": CANCEL_TIMEOUT,
        }
        with pytest.raises(ValueError, match=CONFLICTING_DEFERABLE_WAIT_UNTIL_FINISHED):
            DataflowTemplatedJobStartOperator(**init_kwargs)

    @pytest.mark.db_test
    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook.start_template_dataflow")
    def test_start_with_custom_region(self, dataflow_mock):
        init_kwargs = {
            "task_id": TASK_ID,
            "template": TEMPLATE,
            "dataflow_default_options": {
                "region": TEST_REGION,
            },
            "poll_sleep": POLL_SLEEP,
            "wait_until_finished": True,
            "cancel_timeout": CANCEL_TIMEOUT,
        }
        operator = DataflowTemplatedJobStartOperator(**init_kwargs)
        mock_context = {"task_instance": mock.MagicMock()}
        operator.execute(mock_context)
        assert dataflow_mock.called
        _, kwargs = dataflow_mock.call_args_list[0]
        assert kwargs["variables"]["region"] == TEST_REGION
        assert kwargs["location"] == DEFAULT_DATAFLOW_LOCATION

    @pytest.mark.db_test
    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook.start_template_dataflow")
    def test_start_with_location(self, dataflow_mock):
        init_kwargs = {
            "task_id": TASK_ID,
            "template": TEMPLATE,
            "location": TEST_LOCATION,
            "poll_sleep": POLL_SLEEP,
            "wait_until_finished": True,
            "cancel_timeout": CANCEL_TIMEOUT,
        }
        operator = DataflowTemplatedJobStartOperator(**init_kwargs)
        mock_context = {"task_instance": mock.MagicMock()}
        operator.execute(mock_context)
        assert dataflow_mock.called
        _, kwargs = dataflow_mock.call_args_list[0]
        assert not kwargs["variables"]
        assert kwargs["location"] == TEST_LOCATION


class TestDataflowStartFlexTemplateOperator:
    @pytest.fixture
    def sync_operator(self):
        return DataflowStartFlexTemplateOperator(
            task_id="start_flex_template_streaming_beam_sql",
            body={"launchParameter": TEST_FLEX_PARAMETERS},
            do_xcom_push=True,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            expected_terminal_state=DataflowJobStatus.JOB_STATE_DONE,
        )

    @pytest.fixture
    def deferrable_operator(self):
        return DataflowStartFlexTemplateOperator(
            task_id="start_flex_template_streaming_beam_sql",
            body={"launchParameter": TEST_FLEX_PARAMETERS},
            do_xcom_push=True,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            deferrable=True,
        )

    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute(self, mock_dataflow, sync_operator):
        sync_operator.execute(mock.MagicMock())
        mock_dataflow.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            drain_pipeline=False,
            expected_terminal_state=DataflowJobStatus.JOB_STATE_DONE,
            cancel_timeout=600,
            wait_until_finished=None,
            impersonation_chain=None,
            poll_sleep=10,
        )
        mock_dataflow.return_value.start_flex_template.assert_called_once_with(
            body={"launchParameter": TEST_FLEX_PARAMETERS},
            location=TEST_LOCATION,
            project_id=TEST_PROJECT,
            on_new_job_callback=mock.ANY,
        )

    def test_on_kill(self, sync_operator):
        sync_operator.hook = mock.MagicMock()
        sync_operator.job = {"id": JOB_ID, "projectId": TEST_PROJECT, "location": TEST_LOCATION}
        sync_operator.on_kill()
        sync_operator.hook.cancel_job.assert_called_once_with(
            job_id="test-dataflow-pipeline-id", project_id=TEST_PROJECT, location=TEST_LOCATION
        )

    def test_validation_deferrable_params_raises_error(self):
        init_kwargs = {
            "task_id": "start_flex_template_streaming_beam_sql",
            "body": {"launchParameter": TEST_FLEX_PARAMETERS},
            "do_xcom_push": True,
            "location": TEST_LOCATION,
            "project_id": TEST_PROJECT,
            "wait_until_finished": True,
            "deferrable": True,
        }
        with pytest.raises(ValueError, match=CONFLICTING_DEFERABLE_WAIT_UNTIL_FINISHED):
            DataflowStartFlexTemplateOperator(**init_kwargs)

    @mock.patch(f"{DATAFLOW_PATH}.DataflowStartFlexTemplateOperator.defer")
    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute_with_deferrable_mode(self, mock_hook, mock_defer_method, deferrable_operator):
        deferrable_operator.execute(mock.MagicMock())

        mock_hook.return_value.launch_job_with_flex_template.assert_called_once_with(
            body={"launchParameter": TEST_FLEX_PARAMETERS},
            location=TEST_LOCATION,
            project_id=TEST_PROJECT,
        )
        mock_defer_method.assert_called_once()


class TestDataflowStartYamlJobOperator:
    @pytest.fixture
    def sync_operator(self):
        return DataflowStartYamlJobOperator(
            task_id="start_dataflow_yaml_job_sync",
            job_name="dataflow_yaml_job",
            yaml_pipeline_file="test_file_path",
            append_job_name=False,
            project_id=TEST_PROJECT,
            region=TEST_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            expected_terminal_state=DataflowJobStatus.JOB_STATE_DONE,
        )

    @pytest.fixture
    def deferrable_operator(self):
        return DataflowStartYamlJobOperator(
            task_id="start_dataflow_yaml_job_def",
            job_name="dataflow_yaml_job",
            yaml_pipeline_file="test_file_path",
            append_job_name=False,
            project_id=TEST_PROJECT,
            region=TEST_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            deferrable=True,
            expected_terminal_state=DataflowJobStatus.JOB_STATE_RUNNING,
        )

    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute(self, mock_hook, sync_operator):
        sync_operator.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(
            poll_sleep=sync_operator.poll_sleep,
            drain_pipeline=False,
            impersonation_chain=None,
            cancel_timeout=sync_operator.cancel_timeout,
            expected_terminal_state=DataflowJobStatus.JOB_STATE_DONE,
            gcp_conn_id=GCP_CONN_ID,
        )
        mock_hook.return_value.launch_beam_yaml_job.assert_called_once_with(
            job_name=sync_operator.job_name,
            yaml_pipeline_file=sync_operator.yaml_pipeline_file,
            append_job_name=False,
            options=None,
            jinja_variables=None,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
        )

    @mock.patch(f"{DATAFLOW_PATH}.DataflowStartYamlJobOperator.defer")
    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute_with_deferrable_mode(self, mock_hook, mock_defer_method, deferrable_operator):
        deferrable_operator.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(
            poll_sleep=deferrable_operator.poll_sleep,
            drain_pipeline=False,
            impersonation_chain=None,
            cancel_timeout=deferrable_operator.cancel_timeout,
            expected_terminal_state=DataflowJobStatus.JOB_STATE_RUNNING,
            gcp_conn_id=GCP_CONN_ID,
        )
        mock_defer_method.assert_called_once()

    @mock.patch(f"{DATAFLOW_PATH}.DataflowHook")
    def test_execute_complete_success(self, mock_hook, deferrable_operator):
        expected_result = {"id": JOB_ID}
        mock_context = {"task_instance": mock.MagicMock()}
        actual_result = deferrable_operator.execute_complete(
            context=mock_context,
            event={
                "status": "success",
                "message": "Batch job completed.",
                "job": expected_result,
            },
        )
        assert actual_result == expected_result

    def test_execute_complete_error_status_raises_exception(self, deferrable_operator):
        with pytest.raises(AirflowException, match="Job failed."):
            deferrable_operator.execute_complete(
                context=None, event={"status": "error", "message": "Job failed."}
            )
        with pytest.raises(AirflowException, match="Job was stopped."):
            deferrable_operator.execute_complete(
                context=None, event={"status": "stopped", "message": "Job was stopped."}
            )


class TestDataflowStopJobOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    def test_exec_job_id(self, dataflow_mock):
        self.dataflow = DataflowStopJobOperator(
            task_id=TASK_ID,
            project_id=TEST_PROJECT,
            job_id=JOB_ID,
            poll_sleep=POLL_SLEEP,
            location=TEST_LOCATION,
        )
        """
        Test DataflowHook is created and the right args are passed to cancel_job.
        """
        cancel_job_hook = dataflow_mock.return_value.cancel_job
        mock_context = {"task_instance": mock.MagicMock()}
        self.dataflow.execute(mock_context)
        assert dataflow_mock.called
        cancel_job_hook.assert_called_once_with(
            job_name=None,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            job_id=JOB_ID,
        )

    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    def test_exec_job_name_prefix(self, dataflow_mock):
        self.dataflow = DataflowStopJobOperator(
            task_id=TASK_ID,
            project_id=TEST_PROJECT,
            job_name_prefix=JOB_NAME,
            poll_sleep=POLL_SLEEP,
            location=TEST_LOCATION,
        )
        """
        Test DataflowHook is created and the right args are passed to cancel_job
        and is_job_dataflow_running.
        """
        is_job_running_hook = dataflow_mock.return_value.is_job_dataflow_running
        cancel_job_hook = dataflow_mock.return_value.cancel_job
        mock_context = {"task_instance": mock.MagicMock()}
        self.dataflow.execute(mock_context)
        assert dataflow_mock.called
        is_job_running_hook.assert_called_once_with(
            name=JOB_NAME,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
        )
        cancel_job_hook.assert_called_once_with(
            job_name=JOB_NAME,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            job_id=None,
        )


class TestDataflowCreatePipelineOperator:
    @pytest.fixture
    def create_operator(self):
        """
        Creates a mock create datapipeline operator to be used in testing.
        """
        return DataflowCreatePipelineOperator(
            task_id="test_create_datapipeline",
            body=TEST_PIPELINE_BODY,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
        )

    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    def test_execute(self, mock_hook, create_operator):
        """
        Test that operator creates and calls the Dataflow Data Pipeline hook with the correct parameters
        """
        create_operator.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id="test_gcp_conn_id",
            impersonation_chain=None,
        )

        mock_hook.return_value.create_data_pipeline.assert_called_once_with(
            project_id=TEST_PROJECT, body=TEST_PIPELINE_BODY, location=TEST_LOCATION
        )

    def test_body_invalid(self, sdk_connection_not_found):
        """
        Test that if the operator is not passed a Request Body, an AirflowException is raised
        """
        init_kwargs = {
            "task_id": "test_create_datapipeline",
            "body": {},
            "project_id": TEST_PROJECT,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowCreatePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_projectid_invalid(self):
        """
        Test that if the operator is not passed a Project ID, an AirflowException is raised
        """
        init_kwargs = {
            "task_id": "test_create_datapipeline",
            "body": TEST_PIPELINE_BODY,
            "project_id": None,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowCreatePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_location_invalid(self):
        """
        Test that if the operator is not passed a location, an AirflowException is raised
        """
        init_kwargs = {
            "task_id": "test_create_datapipeline",
            "body": TEST_PIPELINE_BODY,
            "project_id": TEST_PROJECT,
            "location": None,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowCreatePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_response_invalid(self, sdk_connection_not_found):
        """
        Test that if the Response Body contains an error message, an AirflowException is raised
        """
        init_kwargs = {
            "task_id": "test_create_datapipeline",
            "body": {"name": TEST_PIPELINE_NAME, "error": "Testing that AirflowException is raised"},
            "project_id": TEST_PROJECT,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowCreatePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    def test_response_409(self, mock_hook, create_operator):
        """
        Test that if the Pipeline already exists, the operator does not fail and retrieves existed Pipeline
        """
        mock_hook.return_value.create_data_pipeline.side_effect = HttpError(
            resp=httplib2.Response({"status": "409"}), content=b"content"
        )
        create_operator.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id="test_gcp_conn_id",
            impersonation_chain=None,
        )

        mock_hook.return_value.get_data_pipeline.assert_called_once_with(
            project_id=TEST_PROJECT, pipeline_name=TEST_PIPELINE_NAME, location=TEST_LOCATION
        )


class TestDataflowRunPipelineOperator:
    @pytest.fixture
    def run_operator(self):
        """
        Create a DataflowRunPipelineOperator instance with test data
        """
        return DataflowRunPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=TEST_PIPELINE_NAME,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
        )

    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    def test_execute(self, data_pipeline_hook_mock, run_operator):
        """
        Test Run Operator execute with correct parameters
        """
        run_operator.execute(mock.MagicMock())
        data_pipeline_hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )

        data_pipeline_hook_mock.return_value.run_data_pipeline.assert_called_once_with(
            pipeline_name=TEST_PIPELINE_NAME,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
        )

    def test_invalid_data_pipeline_name(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Run Operator is not given a data pipeline name.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": None,
            "project_id": TEST_PROJECT,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowRunPipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_invalid_project_id(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Run Operator is not given a project ID.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": TEST_PIPELINE_NAME,
            "project_id": None,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowRunPipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_invalid_location(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Run Operator is not given a location.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": TEST_PIPELINE_NAME,
            "project_id": TEST_PROJECT,
            "location": None,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowRunPipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_invalid_response(self):
        """
        Test that AirflowException is raised if Run Operator fails execution and returns error.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "data_pipeline_name": TEST_PIPELINE_NAME,
            "project_id": TEST_PROJECT,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises((TypeError, AirflowException), match="missing keyword argument"):
            DataflowRunPipelineOperator(**init_kwargs).execute(mock.MagicMock()).return_value = {
                "error": {"message": "example error"}
            }

    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    def test_response_404(self, mock_hook, run_operator):
        """
        Test that if the Pipeline does not exist, the operator raise AirflowException
        """
        mock_hook.return_value.run_data_pipeline.side_effect = HttpError(
            resp=httplib2.Response({"status": "404"}), content=b"content"
        )
        with pytest.raises(AirflowException):
            run_operator.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id="test_gcp_conn_id",
            impersonation_chain=None,
        )


class TestDataflowDeletePipelineOperator:
    @pytest.fixture
    def run_operator(self):
        """
        Create a DataflowDeletePipelineOperator instance with test data
        """
        return DataflowDeletePipelineOperator(
            task_id=TASK_ID,
            pipeline_name=TEST_PIPELINE_NAME,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
        )

    @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook")
    # @mock.patch("airflow.providers.google.cloud.operators.dataflow.DataflowHook.delete_data_pipeline")
    def test_execute(self, data_pipeline_hook_mock, run_operator):
        """
        Test Delete Operator execute with correct parameters
        """
        data_pipeline_hook_mock.return_value.delete_data_pipeline.return_value = None
        run_operator.execute(mock.MagicMock())
        data_pipeline_hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )

        data_pipeline_hook_mock.return_value.delete_data_pipeline.assert_called_once_with(
            pipeline_name=TEST_PIPELINE_NAME,
            project_id=TEST_PROJECT,
            location=TEST_LOCATION,
        )

    def test_invalid_data_pipeline_name(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Delete Operator is not given a data pipeline name.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": None,
            "project_id": TEST_PROJECT,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowDeletePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_invalid_project_id(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Delete Operator is not given a project ID.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": TEST_PIPELINE_NAME,
            "project_id": None,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowDeletePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_invalid_location(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Delete Operator is not given a location.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": TEST_PIPELINE_NAME,
            "project_id": TEST_PROJECT,
            "location": None,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowDeletePipelineOperator(**init_kwargs).execute(mock.MagicMock())

    def test_invalid_response(self, sdk_connection_not_found):
        """
        Test that AirflowException is raised if Delete Operator fails execution and returns error.
        """
        init_kwargs = {
            "task_id": TASK_ID,
            "pipeline_name": TEST_PIPELINE_NAME,
            "project_id": TEST_PROJECT,
            "location": TEST_LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
        }
        with pytest.raises(AirflowException):
            DataflowDeletePipelineOperator(**init_kwargs).execute(mock.MagicMock()).return_value = {
                "error": {"message": "example error"}
            }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that uses Google Cloud Batch Operators.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud import batch_v1

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.cloud_batch import (
    CloudBatchDeleteJobOperator,
    CloudBatchListJobsOperator,
    CloudBatchListTasksOperator,
    CloudBatchSubmitJobOperator,
)
from airflow.providers.standard.operators.python import PythonOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "cloud_batch"
REGION = "us-central1"

job_name_prefix = "batch"
job1_name = f"{job_name_prefix}-{DAG_ID}-{ENV_ID}-1".replace("_", "-")
job2_name = f"{job_name_prefix}-{DAG_ID}-{ENV_ID}-2".replace("_", "-")

list_jobs_task_name = "list-jobs"
list_tasks_task_name = "list-tasks"


def _assert_jobs(ti):
    job_list = ti.xcom_pull(list_jobs_task_name)
    job_names_str = ""

    if job_list:
        for job in job_list:
            job_names_str += job["name"].split("/")[-1] + " "

    assert job1_name in job_names_str
    assert job2_name in job_names_str


def _assert_tasks(ti):
    task_list = ti.xcom_pull(list_tasks_task_name)

    assert len(task_list) == 2
    assert "tasks/0" in task_list[0]["name"]
    assert "tasks/1" in task_list[1]["name"]


# [START howto_operator_batch_job_creation]
def _create_job():
    runnable = batch_v1.Runnable()
    runnable.container = batch_v1.Runnable.Container()
    runnable.container.image_uri = "gcr.io/google-containers/busybox"
    runnable.container.entrypoint = "/bin/sh"
    runnable.container.commands = [
        "-c",
        "echo Hello world! This is task ${BATCH_TASK_INDEX}.\
          This job has a total of ${BATCH_TASK_COUNT} tasks.",
    ]

    task = batch_v1.TaskSpec()
    task.runnables = [runnable]

    resources = batch_v1.ComputeResource()
    resources.cpu_milli = 2000
    resources.memory_mib = 16
    task.compute_resource = resources
    task.max_retry_count = 2

    group = batch_v1.TaskGroup()
    group.task_count = 2
    group.task_spec = task
    policy = batch_v1.AllocationPolicy.InstancePolicy()
    policy.machine_type = "e2-standard-4"
    instances = batch_v1.AllocationPolicy.InstancePolicyOrTemplate()
    instances.policy = policy
    allocation_policy = batch_v1.AllocationPolicy()
    allocation_policy.instances = [instances]

    job = batch_v1.Job()
    job.task_groups = [group]
    job.allocation_policy = allocation_policy
    job.labels = {"env": "testing", "type": "container"}

    job.logs_policy = batch_v1.LogsPolicy()
    job.logs_policy.destination = batch_v1.LogsPolicy.Destination.CLOUD_LOGGING

    return job


# [END howto_operator_batch_job_creation]


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "batch"],
) as dag:
    # [START howto_operator_batch_submit_job]
    submit1 = CloudBatchSubmitJobOperator(
        task_id="submit-job1",
        project_id=PROJECT_ID,
        region=REGION,
        job_name=job1_name,
        job=_create_job(),
        dag=dag,
        deferrable=False,
    )
    # [END howto_operator_batch_submit_job]

    # [START howto_operator_batch_submit_job_deferrable_mode]
    submit2 = CloudBatchSubmitJobOperator(
        task_id="submit-job2",
        project_id=PROJECT_ID,
        region=REGION,
        job_name=job2_name,
        job=batch_v1.Job.to_dict(_create_job()),
        dag=dag,
        deferrable=True,
    )
    # [END howto_operator_batch_submit_job_deferrable_mode]

    # [START howto_operator_batch_list_tasks]
    list_tasks = CloudBatchListTasksOperator(
        task_id=list_tasks_task_name, project_id=PROJECT_ID, region=REGION, job_name=job1_name, dag=dag
    )
    # [END howto_operator_batch_list_tasks]

    assert_tasks = PythonOperator(task_id="assert-tasks", python_callable=_assert_tasks, dag=dag)

    # [START howto_operator_batch_list_jobs]
    list_jobs = CloudBatchListJobsOperator(
        task_id=list_jobs_task_name,
        project_id=PROJECT_ID,
        region=REGION,
        limit=10,
        filter=f"name:projects/{PROJECT_ID}/locations/{REGION}/jobs/{job_name_prefix}*",
        dag=dag,
    )
    # [END howto_operator_batch_list_jobs]

    get_name = PythonOperator(task_id="assert-jobs", python_callable=_assert_jobs, dag=dag)

    # [START howto_operator_delete_job]
    delete_job1 = CloudBatchDeleteJobOperator(
        task_id="delete-job1",
        project_id=PROJECT_ID,
        region=REGION,
        job_name=job1_name,
        dag=dag,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_delete_job]

    delete_job2 = CloudBatchDeleteJobOperator(
        task_id="delete-job2",
        project_id=PROJECT_ID,
        region=REGION,
        job_name=job2_name,
        dag=dag,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    ([submit1, submit2] >> list_tasks >> assert_tasks >> list_jobs >> get_name >> [delete_job1, delete_job2])

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.marketing_platform.sensors.campaign_manager import (
    GoogleCampaignManagerReportSensor,
)

API_VERSION = "v4"
GCP_CONN_ID = "google_cloud_default"


class TestGoogleCampaignManagerDeleteReportOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.sensors.campaign_manager.GoogleCampaignManagerHook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.sensors.campaign_manager.BaseSensorOperator")
    def test_execute(self, mock_base_op, hook_mock):
        profile_id = "PROFILE_ID"
        report_id = "REPORT_ID"
        file_id = "FILE_ID"

        hook_mock.return_value.get_report.return_value = {"status": "REPORT_AVAILABLE"}

        op = GoogleCampaignManagerReportSensor(
            profile_id=profile_id,
            report_id=report_id,
            file_id=file_id,
            api_version=API_VERSION,
            task_id="test_task",
        )
        result = op.poke(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=None,
        )
        hook_mock.return_value.get_report.assert_called_once_with(
            profile_id=profile_id, report_id=report_id, file_id=file_id
        )
        assert result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

import google.cloud.exceptions
from google.api_core.exceptions import AlreadyExists
from google.cloud.logging_v2.types import LogSink

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_logging import CloudLoggingHook
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from google.protobuf.field_mask_pb2 import FieldMask

    from airflow.providers.common.compat.sdk import Context


def _validate_inputs(obj, required_fields: list[str]) -> None:
    """Validate that all required fields are present on self."""
    missing = [field for field in required_fields if not getattr(obj, field, None)]
    if missing:
        raise AirflowException(
            f"Required parameters are missing: {missing}. These must be passed as keyword parameters."
        )


def _get_field(obj, field_name):
    """Supports both dict and protobuf-like objects."""
    if isinstance(obj, dict):
        return obj.get(field_name)
    return getattr(obj, field_name, None)


class CloudLoggingCreateSinkOperator(GoogleCloudBaseOperator):
    """
    Creates a Cloud Logging export sink in a GCP project.

    This operator creates a sink that exports log entries from Cloud Logging
    to destinations like Cloud Storage, BigQuery, or Pub/Sub.

    :param project_id: Required. ID of the Google Cloud project where the sink will be created.
    :param sink_config: Required. The full sink configuration as a dictionary or a LogSink object.
        See: https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks
    :param unique_writer_identity: If True, creates a unique service account for the sink.
        If False, uses the default Google-managed service account.
    :param gcp_conn_id: Optional. The connection ID used to connect to Google Cloud. Defaults to "google_cloud_default".
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "sink_config",
        "gcp_conn_id",
        "impersonation_chain",
        "unique_writer_identity",
    )

    def __init__(
        self,
        project_id: str,
        sink_config: dict | LogSink,
        unique_writer_identity: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.project_id = project_id
        self.sink_config = sink_config
        self.unique_writer_identity = unique_writer_identity
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict[str, Any]:
        """Execute the operator."""
        _validate_inputs(self, required_fields=["project_id", "sink_config"])
        hook = CloudLoggingHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)

        try:
            self.log.info(
                "Creating log sink '%s' in project '%s'",
                _get_field(self.sink_config, "name"),
                self.project_id,
            )
            self.log.info("Destination: %s", _get_field(self.sink_config, "destination"))

            response = hook.create_sink(
                sink=self.sink_config,
                unique_writer_identity=self.unique_writer_identity,
                project_id=self.project_id,
            )

            self.log.info("Log sink created successfully: %s", response.name)

            if self.unique_writer_identity and hasattr(response, "writer_identity"):
                self.log.info("Writer identity: %s", response.writer_identity)
                self.log.info("Remember to grant appropriate permissions to the writer identity")

            return LogSink.to_dict(response)

        except AlreadyExists:
            self.log.info(
                "Already existed log sink, sink_name=%s, project_id=%s",
                _get_field(self.sink_config, "name"),
                self.project_id,
            )
            existing_sink = hook.get_sink(
                sink_name=_get_field(self.sink_config, "name"), project_id=self.project_id
            )
            return LogSink.to_dict(existing_sink)

        except google.cloud.exceptions.GoogleCloudError as e:
            self.log.error("An error occurred. Exiting.")
            raise e


class CloudLoggingDeleteSinkOperator(GoogleCloudBaseOperator):
    """
    Deletes a Cloud Logging export sink from a GCP project.

    :param sink_name: Required. Name of the sink to delete.
    :param project_id: Required. The ID of the Google Cloud project.
    :param gcp_conn_id: Optional. The connection ID to use for connecting to Google Cloud.
        Defaults to "google_cloud_default".
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = ("sink_name", "project_id", "gcp_conn_id", "impersonation_chain")

    def __init__(
        self,
        sink_name: str,
        project_id: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sink_name = sink_name
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        """Execute the operator."""
        _validate_inputs(self, ["sink_name", "project_id"])
        hook = CloudLoggingHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)

        try:
            self.log.info("Deleting log sink '%s' from project '%s'", self.sink_name, self.project_id)
            hook.delete_sink(sink_name=self.sink_name, project_id=self.project_id)
            self.log.info("Log sink '%s' deleted successfully", self.sink_name)

        except google.cloud.exceptions.NotFound as e:
            self.log.error("An error occurred. Not Found.")
            raise e
        except google.cloud.exceptions.GoogleCloudError as e:
            self.log.error("An error occurred. Exiting.")
            raise e


class CloudLoggingUpdateSinkOperator(GoogleCloudBaseOperator):
    """
    Updates an existing Cloud Logging export sink.

    :param project_id: Required. The ID of the Google Cloud project that contains the sink.
    :param sink_name: Required. The name of the sink to update.
    :param sink_config: Required. The updated sink configuration. Can be a dictionary or a
        `google.cloud.logging_v2.types.LogSink` object. Refer to:
        https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks
    :param update_mask: Required. A FieldMask or dictionary specifying which fields of the sink
        should be updated. For example, to update the destination and filter, use:
        `{"paths": ["destination", "filter"]}`.
    :param unique_writer_identity: Optional. When set to True, a new unique service account
        will be created for the sink. Defaults to False.
    :param gcp_conn_id: Optional. The connection ID used to connect to Google Cloud.
        Defaults to "google_cloud_default".
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "sink_name",
        "project_id",
        "update_mask",
        "sink_config",
        "unique_writer_identity",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        project_id: str,
        sink_name: str,
        sink_config: dict | LogSink,
        update_mask: FieldMask | dict,
        unique_writer_identity: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.project_id = project_id
        self.sink_name = sink_name
        self.sink_config = sink_config
        self.update_mask = update_mask
        self.unique_writer_identity = unique_writer_identity
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict[str, Any]:
        """Execute the operator."""
        _validate_inputs(self, ["sink_name", "project_id", "sink_config", "update_mask"])
        hook = CloudLoggingHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)

        try:
            current_sink = hook.get_sink(sink_name=self.sink_name, project_id=self.project_id)
            self.log.info("Current log sink configuration: '%s'.", LogSink.to_dict(current_sink))

            self.log.info("Updating log sink '%s' in project '%s'", self.sink_name, self.project_id)
            if isinstance(self.update_mask, dict) and "paths" in self.update_mask:
                paths = self.update_mask["paths"]
            elif hasattr(self.update_mask, "paths"):
                paths = self.update_mask.paths

            self.log.info("Updating fields: %s", ", ".join(paths))

            response = hook.update_sink(
                sink_name=self.sink_name,
                sink=self.sink_config,
                unique_writer_identity=self.unique_writer_identity,
                project_id=self.project_id,
                update_mask=self.update_mask,
            )
            self.log.info("Log sink updated successfully: %s", response.name)
            return LogSink.to_dict(response)

        except google.cloud.exceptions.NotFound as e:
            self.log.error("An error occurred. Not Found.")
            raise e
        except google.cloud.exceptions.GoogleCloudError as e:
            self.log.error("An error occurred. Exiting.")
            raise e


class CloudLoggingListSinksOperator(GoogleCloudBaseOperator):
    """
    Lists Cloud Logging export sinks in a Google Cloud project.

    :param project_id: Required. The ID of the Google Cloud project to list sinks from.
    :param page_size: Optional. The maximum number of sinks to return per page. Must be greater than 0.
        If None, the server will use a default value.
    :param gcp_conn_id: Optional. The connection ID used to connect to Google Cloud.
        Defaults to "google_cloud_default".
    :param impersonation_chain: Optional. Service account or chained list of accounts to impersonate.
        If a string, the service account must grant the originating account the
        'Service Account Token Creator' IAM role.

        If a sequence, each account in the chain must grant this role to the next.
        The first account must grant it to the originating account (templated).
    """

    template_fields: Sequence[str] = ("project_id", "gcp_conn_id", "impersonation_chain", "page_size")

    def __init__(
        self,
        project_id: str,
        page_size: int | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.project_id = project_id
        self.page_size = page_size
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> list[dict[str, Any]]:
        """Execute the operator."""
        _validate_inputs(self, ["project_id"])

        if self.page_size is not None and self.page_size < 1:
            raise AirflowException("The page_size for the list sinks request must be greater than zero")

        hook = CloudLoggingHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)

        try:
            self.log.info("Listing log sinks in project '%s'", self.project_id)

            sinks = hook.list_sinks(project_id=self.project_id, page_size=self.page_size)

            result = [LogSink.to_dict(sink) for sink in sinks]
            self.log.info("Found %d log sinks", len(result))

            return result

        except google.cloud.exceptions.GoogleCloudError as e:
            self.log.error("An error occurred. Exiting.")
            raise e



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Secret Manager hook."""

from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING

from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.secretmanager_v1 import (
    AccessSecretVersionResponse,
    Secret,
    SecretManagerServiceClient,
    SecretPayload,
    SecretVersion,
)

from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

if TYPE_CHECKING:
    from google.api_core.retry import Retry
    from google.cloud.secretmanager_v1.services.secret_manager_service.pagers import ListSecretsPager


class GoogleCloudSecretManagerHook(GoogleBaseHook):
    """
    Hook for the Google Cloud Secret Manager API.

    See https://cloud.google.com/secret-manager
    """

    def __init__(self, location: str | None = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.location = location

    @cached_property
    def client(self):
        """
        Create a Secret Manager Client.

        :return: Secret Manager client.
        """
        if self.location is not None:
            return SecretManagerServiceClient(
                credentials=self.get_credentials(),
                client_info=CLIENT_INFO,
                client_options={"api_endpoint": f"secretmanager.{self.location}.rep.googleapis.com"},
            )
        return SecretManagerServiceClient(
            credentials=self.get_credentials(),
            client_info=CLIENT_INFO,
        )

    def get_conn(self) -> SecretManagerServiceClient:
        """
        Retrieve the connection to Secret Manager.

        :return: Secret Manager client.
        """
        return self.client

    def _get_parent(self, project_id: str, location: str | None = None) -> str:
        """
        Return parent path.

        :param project_id: Required. ID of the GCP project that owns the job.
        :param location: Optional. Target location.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations

        :return: Parent path.
        """
        _location = location or self.location
        if _location is not None:
            return self.client.common_location_path(project_id, _location)
        return self.client.common_project_path(project_id)

    def _get_secret_path(self, project_id: str, secret_id: str, location: str | None = None) -> str:
        """
        Return secret path.

        :param project_id: Required. ID of the GCP project that owns the job.
        :param secret_id: Required. Secret ID for which path is required.
        :param location: Optional. Target location.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations

        :return: Parent path.
        """
        _location = location or self.location
        if _location is not None:
            # Google's client library does not provide a method to construct regional secret paths, so constructing manually.
            return f"projects/{project_id}/locations/{_location}/secrets/{secret_id}"
        return self.client.secret_path(project_id, secret_id)

    def _get_secret_version_path(
        self, project_id: str, secret_id: str, secret_version: str, location: str | None = None
    ) -> str:
        """
        Return secret version path.

        :param project_id: Required. ID of the GCP project that owns the job.
        :param secret_id: Required. Secret ID for which path is required.
        :param secret_version: Required. Secret version for which path is required.
        :param location: Optional. Target location.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations

        :return: Parent path.
        """
        _location = location or self.location
        if _location is not None:
            # Google's client library does not provide a method to construct regional secret version paths, so constructing manually.
            return (
                f"projects/{project_id}/locations/{_location}/secrets/{secret_id}/versions/{secret_version}"
            )
        return self.client.secret_version_path(project_id, secret_id, secret_version)

    @GoogleBaseHook.fallback_to_default_project_id
    def create_secret(
        self,
        project_id: str,
        secret_id: str,
        secret: dict | Secret | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        location: str | None = None,
    ) -> Secret:
        """
        Create a secret.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/secretmanager/latest/google.cloud.secretmanager_v1.services.secret_manager_service.SecretManagerServiceClient#google_cloud_secretmanager_v1_services_secret_manager_service_SecretManagerServiceClient_create_secret

        :param project_id: Required. ID of the GCP project that owns the job.
            If set to ``None`` or missing, the default project_id from the GCP connection is used.
        :param secret_id: Required. ID of the secret to create.
        :param secret: Optional. Secret to create.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        :param location: Optional. Location where secret should be created. Used for creating regional secret.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations
        :return: Secret object.
        """
        if not secret:
            _secret: dict | Secret = {}
            if (location or self.location) is None:
                _secret["replication"] = {"automatic": {}}
        else:
            _secret = secret

        response = self.client.create_secret(
            request={
                "parent": self._get_parent(project_id, location),
                "secret_id": secret_id,
                "secret": _secret,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Secret Created: %s", response.name)
        return response

    @GoogleBaseHook.fallback_to_default_project_id
    def add_secret_version(
        self,
        project_id: str,
        secret_id: str,
        secret_payload: dict | SecretPayload | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        location: str | None = None,
    ) -> SecretVersion:
        """
        Add a version to the secret.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/secretmanager/latest/google.cloud.secretmanager_v1.services.secret_manager_service.SecretManagerServiceClient#google_cloud_secretmanager_v1_services_secret_manager_service_SecretManagerServiceClient_add_secret_version

        :param project_id: Required. ID of the GCP project that owns the job.
            If set to ``None`` or missing, the default project_id from the GCP connection is used.
        :param secret_id: Required. ID of the secret to create.
        :param secret_payload: Optional. A secret payload.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        :param location: Optional. Location where secret is located. Used for adding version to regional secret.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations
        :return: Secret version object.
        """
        response = self.client.add_secret_version(
            request={
                "parent": self._get_secret_path(project_id, secret_id, location),
                "payload": secret_payload,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Secret version added: %s", response.name)
        return response

    @GoogleBaseHook.fallback_to_default_project_id
    def list_secrets(
        self,
        project_id: str,
        page_size: int = 0,
        page_token: str | None = None,
        secret_filter: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        location: str | None = None,
    ) -> ListSecretsPager:
        """
        List secrets.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/secretmanager/latest/google.cloud.secretmanager_v1.services.secret_manager_service.SecretManagerServiceClient#google_cloud_secretmanager_v1_services_secret_manager_service_SecretManagerServiceClient_list_secrets

        :param project_id: Required. ID of the GCP project that owns the job.
            If set to ``None`` or missing, the default project_id from the GCP connection is used.
        :param page_size: Optional, number of results to return in the list.
        :param page_token: Optional, token to provide to skip to a particular spot in the list.
        :param secret_filter: Optional. Filter string.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        :param location: Optional. The regional secrets stored in the provided location will be listed.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations
        :return: Secret List object.
        """
        response = self.client.list_secrets(
            request={
                "parent": self._get_parent(project_id, location),
                "page_size": page_size,
                "page_token": page_token,
                "filter": secret_filter,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Secrets list obtained")
        return response

    @GoogleBaseHook.fallback_to_default_project_id
    def secret_exists(self, project_id: str, secret_id: str, location: str | None = None) -> bool:
        """
        Check whether secret exists.

        :param project_id: Required. ID of the GCP project that owns the job.
            If set to ``None`` or missing, the default project_id from the GCP connection is used.
        :param secret_id: Required. ID of the secret to find.
        :param location: Optional. Location where secret is expected to be stored regionally.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations
        :return: True if the secret exists, False otherwise.
        """
        secret_filter = f"name:{secret_id}"
        secret_name = self._get_secret_path(project_id, secret_id, location)
        for secret in self.list_secrets(
            project_id=project_id, page_size=100, secret_filter=secret_filter, location=location
        ):
            if secret.name.split("/")[-1] == secret_id:
                self.log.info("Secret %s exists.", secret_name)
                return True
        self.log.info("Secret %s doesn't exists.", secret_name)
        return False

    @GoogleBaseHook.fallback_to_default_project_id
    def access_secret(
        self,
        project_id: str,
        secret_id: str,
        secret_version: str = "latest",
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        location: str | None = None,
    ) -> AccessSecretVersionResponse:
        """
        Access a secret version.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/secretmanager/latest/google.cloud.secretmanager_v1.services.secret_manager_service.SecretManagerServiceClient#google_cloud_secretmanager_v1_services_secret_manager_service_SecretManagerServiceClient_access_secret_version

        :param project_id: Required. ID of the GCP project that owns the job.
            If set to ``None`` or missing, the default project_id from the GCP connection is used.
        :param secret_id: Required. ID of the secret to access.
        :param secret_version: Optional. Version of the secret to access. Default: latest.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        :param location: Optional. Location where secret is stored regionally.  If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used
            For more details : https://cloud.google.com/secret-manager/docs/locations
        :return: Access secret version response object.
        """
        response = self.client.access_secret_version(
            request={
                "name": self._get_secret_version_path(project_id, secret_id, secret_version, location),
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Secret version accessed: %s", response.name)
        return response

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_secret(
        self,
        project_id: str,
        secret_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        location: str | None = None,
    ) -> None:
        """
        Delete a secret.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/secretmanager/latest/google.cloud.secretmanager_v1.services.secret_manager_service.SecretManagerServiceClient#google_cloud_secretmanager_v1_services_secret_manager_service_SecretManagerServiceClient_delete_secret

        :param project_id: Required. ID of the GCP project that owns the job.
            If set to ``None`` or missing, the default project_id from the GCP connection is used.
        :param secret_id: Required. ID of the secret to delete.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        :param location: Optional. Location where secret is stored regionally. If set to ``None`` or missing, the location provided for GoogleCloudSecretHook instantiation is used.
            For more details : https://cloud.google.com/secret-manager/docs/locations
        :return: Access secret version response object.
        """
        name = self._get_secret_path(project_id, secret_id, location)
        self.client.delete_secret(
            request={"name": name},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Secret deleted: %s", name)
        return None
